/*
 * fm.h
 * standard header for fm
 * Greg Lee, 6/01
 */
#include "../config.h"

#include <stdio.h>
#if STDC_HEADERS
# include <stdlib.h>
# if HAVE_STDDEF_H
# include <stddef.h>
# endif
#else
# if HAVE_STDLIB_H
#  include <stdlib.h>
# endif
#endif
#if HAVE_STRING_H
# include <string.h>
#endif
#if HAVE_STRINGS_H
# include <strings.h>
#endif
#if HAVE_UNISTD_H
# include <unistd.h>
#endif

#if HAVE_CTYPE_H
# include <ctype.h>
#endif
#if HAVE_FCNTL_H
# include <fcntl.h>
#endif

#ifndef TRUE
#define TRUE 1
#endif
#ifndef FALSE
#define FALSE 0
#endif

#define COLWIDTH 500
#define SP(w)	(w*500)
#define COL(w)	(((w>=0)? w+250 : w-250)/500)
#define PT(s) (72 * (s) / ps_pitch_sp)
#define COLMUL 8

#define MIDINSERTION 1
#define VCENTER 2
#define TOPINSERTION 3
#define PAGEINSERTION 4
#define FOOTNOTE 5
#define VBOX 6

/* character attributes controlled by font-changing commands */
#define ROMAN  1	/* plain, from \rm or .R */
#define ITALIC 2	/* underlined, from \it or .I */
#define BOLD   3	/* bold, from \bf or .B */
#define SLANTY 4	/* blinking, from \sl or .G */
#define BITALIC 5	/* bold italic, from \bi */
#define SANS    6	/* sans serif, from \ss */
#define TTYPE   7	/* reversed, from \tt or .S */
#define SUPER   8	/* superscripts, from ^ */
#define SUB     9	/* subscripts, from _ */
#define TITLE     10	/* big, from \title */
#define MATH      11	/* reversed, from $ */
#define IPA       12	/* phonetic */
#define IPASUPER  13	/* phonetic superscript */
#define IPASUB    14	/* phonetic subscript */
#define IPAOVER   15	/* phonetic above diacritic */
#define IPAUNDER  16	/* phonetic under diacritic */
#define MUSIX     17	/* musical notes */
#define BOX       18	/* box drawing characters */
#define ZD        19	/* box drawing characters */

#define is_ipa(c)  \
	((c >= 12 && c <= 16))
#define is_roman(c)  \
	((c >= 1 && c <= 10))

#define SO 0
#define SE 1
#define MB 2
#define MD 3
#define MR 4
#define MH 5
#define ME 6
#define US 7
#define UE 8
#define AS 9
#define AE 10
#define AC 11
#define Sf 12
#define CE 13

#define C_RED 1
#define C_GREEN 2
#define C_YELLOW 3
#define C_BLUE 4
#define C_MAGENTA 5
#define C_CYAN 6
#define C_WHITE 7

#define M_BRIGHT (MD << 4)
#define M_REVERSE (MR << 4)
#define M_DIM (MH << 4)
#define M_BLINK (MB << 4)
#define M_UNDERSCORE (US << 4)

extern int
	c_bold,
	c_italic,
	c_math,
	c_ttype,
	c_slanty,
	c_prompt,
	c_super,
	c_sub,
	c_ipa,
	c_roman,
	c_bitalic,
	c_sans,
	c_musix;

/* Escape characters to communicate special orders to the
 * PostScript interpretive code in ps.c .
 */
#define is_ctl(c) \
	(c <= ' ' || (c >= 0x99 && c < 0xa0))
	/* next 2 chars are unicode wide char */
#define UNI_ESC 0x99
	/* next char is literal */
#define LIT_ESC 0x9a
	/* tab to column */
#define COL_ESC 0x9b
	/* kern */
#define KERN_ESC 0x9c
	/* remember a node position */
#define NODE_ESC 0x9d
	/* issue PS code */
#define PS_ESC 0x9e
	/* draw a line replacing following word */
#define HRULE_ESC 0x9f

#define MAXTL 500
extern struct tline {
	int id1, id2, status;
	int hpos1, hpos2, vpos1, vpos2;
	int left1, left2, right1, right2;
	int named1, named2;
	int empty1, empty2;
	int l, mid, col;
	char attrib[26];
} treeline[MAXTL];

extern unsigned char other[128][6];
extern unsigned char math_other[128][6];
extern int roman_other[128][6];

extern int
        parindent,
        parskip,
        lskip,
        rskip,
	baselineskip,
	pointsize,
        hsize,
        hsize_sp,
        vsize,
	hlevel,
	tabsize,
	tabid,
	outdev,
	hoffset,
	lmargin,
	pageno,
	nopagenumbers,
	pageattop,
	pagestretch,
        offset_sp,
        offset,
        bfont;		/* print with ? */

extern char pageno_str[];
extern void fix_pageno();
extern int romanize(int num, char *s);
extern int  dimen(char *s, int horizontal, int cansign, int wantpoints);
extern int  dimen_sp(char *s, int horizontal, int cansign);
extern void calculate_offset();
extern void calculate_ps_page_length();

#define MAXLINE 1024
#define MAXWORD 1024
#define STACKDEPTH 256
#define MAXTB 24

extern unsigned char *tabline[MAXTB];

/* global state of formatter */
extern char
        iflag,		/* preserving left indentation? */
        stiflag,	/* user set -i flag? */
        staflag,	/* user set -a flag? */
        jflag,		/* requested justification with -j? */
        uflag,		/* no request uncolored output with -u? */
        xflag,		/* requested use TeX commands w. -x? */
        yflag,		/* annotate paragraphs? */
        zflag,		/* debug? */
        mflag,		/* requested use man commands w. -m? */
        nflag,		/* requested no color? */
        oflag,		/* requested overstrike? */
        cflag,		/* requested format C code w. -c? */
        fflag,		/* no pagination? */
        aflag,		/* no show typeface changes with -u? */
        kflag,		/* unicode output? */
        alwayskflag,	/* console always unicode output? */
        gflag,		/* unicode input? */
        vflag,		/* verbose */
        qtflag,		/* saw odd " in processing .RB etc.? */
        ceflag,		/* center next output line? */
        raflag,		/* right adjust next output line? */
        laflag,		/* left adjust next output line? */
        obflag,		/* not concatenating? */
        obsflag,	/* ignore user spacing? */
        liflag,		/* not concatenating? */
        njflag,		/* not currently justifying? */
        mmode,		/* in math mode? */
        rhmode,		/* in restricted horizontal mode? */
        rvmode,		/* in restricted vertical mode? */
	psflag,		/* generating PostScript? */
        lhpenalty,	/* last line ended in hyphen? */
	treeflag,	/* tree in progress */
	boxflag;	/* use BOX font for box chars */

extern int cols, cols_sp;
extern int in_is_tty;
extern int out_is_tty;
extern unsigned char isoprint[256];
extern unsigned char roman_diacritic[256];
extern int hyphenate(unsigned char *word, unsigned char *aword, int len, int hn, int room);
extern int fmcfg();
extern void shipline(unsigned char *oline, unsigned char * aline,
	int len, int tlen, int spaces, int vskip, int vglue, int pointskip, int underscoring, int penalty,
	int insertion);
extern void windup();
extern void terminit();
extern void ttyinit();
extern void devinit();
extern void puttc(int ch);
extern void putts(unsigned char *s);
extern void putword();
extern void putline();
extern int  endofsentence(unsigned char *p, int plen);
extern void scantree(char *base, int count, int tree_indent);

#ifndef STANDALONETREE
extern void pop();
extern void push(unsigned char attr);
extern unsigned char attribute;
extern unsigned char ipa_attribute;
#endif

extern unsigned char glyph[128];

extern struct hboxtype {
	unsigned char *ch, *at;
	int atom, len, tlen, vskip, vfill, vglue, pskip, spaces, us, penalty, tb;
} *hbox[];

#define PS_tilde	 3
#define PS_endash	 128
#define PS_emdash	 129
#define PS_quotedblleft	 130
#define PS_quotedblright   131
#define PS_fi		 132
#define PS_fl		 133
#define PS_ellipsis	 134
#define PS_hungarumlaut	 5
#define PS_grave	 0
#define PS_undergrave	 14
#define PS_circumflex	 2
#define PS_breve	 8
#define PS_underbreve	 16
#define PS_underline	 17
#define PS_stroke	 18
#define PS_slash	 19
#define PS_caron	 7
#define PS_dotaccent	 10
#define PS_underdot	 13
#define PS_ogonek	 12
#define PS_ring		 6
#define PS_dotlessi	 25
#define PS_periodcentered   144
#define PS_lslash	 145
#define PS_Lslash	 146
#define PS_scaron	 147
#define PS_Scaron	 148
#define PS_zcaron	 149
#define PS_Zcaron	 150
#define PS_quotesingle	 151
#define PS_oe		 152
#define PS_OE		 153
#define PS_dagger	 154
#define PS_daggerdbl	 155
#define PS_florin	 156
#define PS_perthousand	 157
#define PS_dieresis	 4
#define PS_macron	 9
#define PS_acute	 1
#define PS_underacute	 15
#define PS_cedilla	 11

#define WORDSPACE	250
#define SENTENCESPACE	500
#define COMMASPACE	375

#ifndef STANDALONETREE
extern unsigned char let[3][256];
extern int numcopies;
extern int ps_page;
extern float ps_pointsize;
extern int ps_pointsize10;
extern float ps_pitch;
extern int ps_pitch_sp;
extern int ps_vpitch_sp;
extern int ps_baselineskip;
extern int ps_baselineskip10;
extern int ps_voffset;

extern void job();
extern void endjob();
extern void pagesetup(int pn);
extern void pagefinish();
extern int pnwidth(unsigned char *s, unsigned char *a, int n);
extern int pwidth(unsigned char *s, unsigned char at, int n);
extern void setfontnames();
extern void ps_outline(int pline, int wh, int i);
#endif

extern int is_combining(unsigned int u);
extern int uni_out(int ch, unsigned char a, int diacritic, unsigned char da);
extern int ipa_width[256];
extern int revtab[][4];
extern int dotab(int c, int tb);
extern int uni_out(int ch, unsigned char a, int diacritic, unsigned char da);

