/*
 * ps.c
 * generate PostScript output for fm
 * Greg Lee, 7/01
 * 
 * Some PostScript code procedures come from GNU enscript by
 * Markku Rossi <mtr@iki.fi> <http://www.iki.fi/~mtr/>.
 */
#include <stdio.h>
#include "fm.h"

#define DEL 127

extern int hv_kern[][3];
extern int tib_kern[][3];
extern int tibi_kern[][3];
extern int tii_kern[][3];
extern int tir_kern[][3];
extern int ipa_kern[][3];

extern int hv_width[256];
extern int hv_y1[256];
extern int hv_y2[256];
extern int courier_kern[][3];
extern int courier_width[256];
extern int courier_y1[256];
extern int courier_y2[256];
extern int tir_width[256];
extern int tii_width[256];
extern int tib_width[256];
extern int tibi_width[256];
extern int tir_y1[256];
extern int tii_y1[256];
extern int tib_y1[256];
extern int tibi_y1[256];
extern int tir_y2[256];
extern int tii_y2[256];
extern int tib_y2[256];
extern int tibi_y2[256];
extern int musix_width[256];
extern int ipa_width[256];
extern int ipa_true_width[256];
extern int ipa_y1[256];
extern int ipa_y2[256];

extern int pob_kern[][3];
extern int pobi_kern[][3];
extern int poi_kern[][3];
extern int por_kern[][3];
extern int por_width[256];
extern int poi_width[256];
extern int pob_width[256];
extern int pobi_width[256];
extern int por_y1[256];
extern int poi_y1[256];
extern int pob_y1[256];
extern int pobi_y1[256];
extern int por_y2[256];
extern int poi_y2[256];
extern int pob_y2[256];
extern int pobi_y2[256];

extern int ncbi_kern[][3];
extern int ncb_kern[][3];
extern int nci_kern[][3];
extern int ncr_kern[][3];
extern int bkd_kern[][3];
extern int bkdi_kern[][3];
extern int bkl_kern[][3];
extern int bkli_kern[][3];

extern int ncb_width[256];
extern int ncbi_width[256];
extern int nci_width[256];
extern int ncr_width[256];
extern int bkd_width[256];
extern int bkdi_width[256];
extern int bkl_width[256];
extern int bkli_width[256];
extern int ncb_y1[256];
extern int ncbi_y1[256];
extern int nci_y1[256];
extern int ncr_y1[256];
extern int bkd_y1[256];
extern int bkdi_y1[256];
extern int bkl_y1[256];
extern int bkli_y1[256];
extern int ncb_y2[256];
extern int ncbi_y2[256];
extern int nci_y2[256];
extern int ncr_y2[256];
extern int bkd_y2[256];
extern int bkdi_y2[256];
extern int bkl_y2[256];
extern int bkli_y2[256];

extern int agd_kern[][3];
extern int agdo_kern[][3];
extern int agw_kern[][3];
extern int agwo_kern[][3];
extern int agd_width[256];
extern int agdo_width[256];
extern int agw_width[256];
extern int agwo_width[256];
extern int agd_y1[256];
extern int agdo_y1[256];
extern int agw_y1[256];
extern int agwo_y1[256];
extern int agd_y2[256];
extern int agdo_y2[256];
extern int agw_y2[256];
extern int agwo_y2[256];

extern int zd_width[256];
extern int sym_width[256];

extern int penalty_line[];
extern char frenchspacing;

int numcopies = 1;
int ps_page = 0;
float ps_pointsize = -2.0;
int ps_pointsize10 = 80;
float ps_pitch = 15.0;
int ps_pitch_sp = 9000;
int ps_vpitch_sp = 4000;
int ps_baselineskip = -1;
int ps_baselineskip10 = -1;
int ps_voffset = 0;

struct tline treeline[MAXTL];

static char *fontname[20];
static int fontusage[20];
static int *roman_width = courier_width;
static int *italic_width = courier_width;
static int *bold_width = courier_width;
static int *slanty_width = courier_width;
/*static int *math_width = courier_width;*/
/*static int *ttype_width = courier_width;*/
extern int ttype_width[256];
extern int ttype_y1[256];
extern int ttype_y2[256];
extern int ttype_kern[][3];
static int *sans_width = courier_width;
static int *bitalic_width = courier_width;
static int *default_width = courier_width;

void setfontnames();

void job();
void endjob();
void pagesetup(int pn);
void pagefinish();

extern int dotab(int c, int tb);
extern void xipa();
#ifdef USEMUSIX
extern void musix();
#endif

static void comments();
static void prolog();
static void trailer(int total);
static void encoding();
static void setup();
static void setfont(unsigned char a);
static void move_to(int this_x, int this_y);

#define TOP_RESERVE 24
#define LEFT_RESERVE 24
#define Letter_Height 792
#define Letter_Width 612
#define Paper_Edge (Letter_Width - LEFT_RESERVE)

static int curr_a = ROMAN;
static char ptsize[40];
static char superptsize[40];
static char ttypeptsize[40];
static char titleptsize[40];
static char it_angle[40];
static char condensedsize[40];

void job() {
	static int job_started = 0;
	int supersize;
	int titlesize;
	int anglesize;
	int narrowsize;
	int ttypesize, rxheight, ttype_height, i;

	if (job_started) return;

	if (ps_pointsize10 < 10) ps_pointsize10 = 120;
	/*if (ps_voffset < -72) ps_voffset = -72;*/
	if (ps_pointsize10 % 10) sprintf(ptsize,"%d.%d",
		 ps_pointsize10/10, ps_pointsize10 % 10);
	else sprintf(ptsize,"%d", ps_pointsize10/10);
	supersize = 3 * ps_pointsize10 / 4;
	if (supersize % 10) sprintf(superptsize,"%d.%d",
		 supersize/10, supersize % 10);
	else sprintf(superptsize,"%d", supersize/10);
	titlesize = 2 * ps_pointsize10;
	sprintf(titleptsize,"%d", titlesize/10);
	anglesize = 267949 * ps_pointsize10 / 1000000;
	if (anglesize % 10) sprintf(it_angle,"%d.%d",
		 anglesize/10, anglesize % 10);
	else sprintf(it_angle,"%d", anglesize/10);

	ttype_height = ttype_y2['x'];

	switch (bfont) {
    		case 0: rxheight = tir_y2['x']; break;
		case 1: rxheight = por_y2['x']; break;
		case 2: rxheight = bkl_y2['x']; break;
		case 3: rxheight = ncr_y2['x']; break;
		case 4: rxheight = agw_y2['x']; break;
		default: rxheight = ttype_height; break;
	}
	ttypesize = ps_pointsize10 * rxheight / ttype_height;
	sprintf(ttypeptsize,"%d.%d",
		 ttypesize/10, ttypesize % 10);

	narrowsize = 4 * ps_pointsize10 * rxheight / 5 / ttype_height;

	if (narrowsize % 10) sprintf(condensedsize,"%d.%d",
		 narrowsize/10, narrowsize % 10);
	else sprintf(condensedsize,"%d", narrowsize/10);

	for (i = 0; i < 256; i++) {
		ttype_width[i] = (ttype_width[i] * rxheight) / ttype_height;
		ttype_y1[i] = (ttype_y1[i] * rxheight) / ttype_height;
		ttype_y2[i] = (ttype_y2[i] * rxheight) / ttype_height;
	}


	comments();
	prolog();
	setup();
	job_started = 1;
}

void endjob() {
	trailer(ps_page); /* total pages */
}

static void comments() {
	extern char *myfile, *myname;
#ifdef HAVE_STRFTIME
	extern char today_str[20], time_str[20];
#endif

	printf("%%!PS-Adobe-3.0\n");
	printf("%%%%BoundingBox: %d %d %d %d\n",
		LEFT_RESERVE, TOP_RESERVE,
		Letter_Width - LEFT_RESERVE,
		Letter_Height - TOP_RESERVE);
	if (myfile) printf("%%%%Title: %s\n", myfile);
	else printf("%%%%Title: Fm Output\n");
	if (myname) printf("%%%%For: %s\n", myname);
	else printf("%%%%For: Greg Lee\n");
	printf("%%%%Creator: Fm 1.0\n");
#ifdef HAVE_STRFTIME
	printf("%%%%CreationDate: %s %s\n", today_str, time_str);
#else
	printf("%%%%CreationDate: Sat Jan  4 14:15:40 2002\n");
#endif
	printf("%%%%Orientation: Portrait\n");
	printf("%%%%Pages: (atend)\n");
	printf("%%%%DocumentMedia: Letter 612 792 0 () ()\n");
	printf("%%%%DocumentNeededResources: (atend)\n");
	printf("%%%%EndComments\n");
}

static void prolog() {

	printf("%%%%BeginProlog\n");
	printf("%%%%BeginResource: procset FM-Prolog 1.0 1\n");
	printf("%%\n");

	printf("%% Procedures.\n");
	printf("%%\n");
	printf("\n");

	printf("/_S {	%% save current state\n");
	printf("  /_s save def\n");
	printf("} def\n");

	printf("/_R {	%% restore from saved state\n");
	printf("  _s restore\n");
	printf("} def\n");
	printf("\n");

	printf("/S {	%% showpage protecting gstate\n");
	printf("  gsave\n");
	printf("  showpage\n");
	printf("  grestore\n");
	printf("} bind def\n");
	printf("\n");

	printf("/MF {	%% fontname newfontname -> -	make a new encoded font\n");
	printf("  /newfontname exch def\n");
	printf("  /fontname exch def\n");
	printf("\n");
	printf("  /fontdict fontname findfont def\n");
	printf("  /newfont fontdict maxlength dict def\n");
	printf("\n");
	printf("  fontdict {\n");
	printf("    exch\n");
	printf("    dup /FID eq {\n");
	printf("      %% skip FID pair\n");
	printf("      pop pop\n");
	printf("    } {\n");
	printf("      %% copy to the new font dictionary\n");
	printf("      exch newfont 3 1 roll put\n");
	printf("    } ifelse\n");
	printf("  } forall\n");
	printf("\n");
	printf("  newfont /FontName newfontname put\n");
	printf("\n");
	printf("  %% insert only valid encoding vectors\n");
	printf("  encoding_vector length 256 eq {\n");
	printf("    newfont /Encoding encoding_vector put\n");
	printf("  } if\n");
	printf("\n");
	printf("  newfontname newfont definefont pop\n");
	printf("} def\n");
	printf("\n");

	printf("/SF { %% fontname width height -> -	set a new font\n");
	printf("  /slant exch def\n");
	printf("  /height exch def\n");
	printf("  /width exch def\n");
	printf("\n");
	printf("  findfont\n");
	printf("  [width 0 slant height 0 0] makefont setfont\n");
	printf("} def\n");
	printf("\n");

	printf("/M {moveto} bind def\n");
	printf("/s {show} bind def\n");
	printf("\n");

	printf("/K {\n");
	printf("  { pop pop 0 rmoveto } exch kshow\n");
	printf("} bind def\n");
	printf("\n");

	printf("/KV {\n");
	printf("  { pop pop rmoveto } exch kshow\n");
	printf("} bind def\n");
	printf("\n");

	printf("%% Page prefeed\n");
	printf("/page_prefeed {		%% bool -> -\n");
	printf("  statusdict /prefeed known {\n");
	printf("    statusdict exch /prefeed exch put\n");
	printf("  } {\n");
	printf("    pop\n");
	printf("  } ifelse\n");
	printf("} def\n");
	printf("\n");

	printf("%% Check PostScript language level.\n");
	printf("/languagelevel where {\n");
	printf("  pop /gs_languagelevel languagelevel def\n");
	printf("} {\n");
	printf("  /gs_languagelevel 1 def\n");
	printf("} ifelse\n");
	printf("%%%%EndResource\n");

	encoding(); /* 8859-1 */

	printf("%%%%BeginResource: procset Tipa-Font 1.6 1\n");
	xipa();
	printf("%%%%EndResource\n");
#ifdef USEMUSIX
	printf("%%%%BeginResource: procset Musix-Font 1.6 1\n");
	musix();
	printf("%%%%EndResource\n");
#endif

	printf("%%%%EndProlog\n");
}


static void encoding() {

	printf("%%%%BeginResource: procset Enscript-Encoding-88591 1.6 1\n");
	printf("/encoding_vector [\n");
	printf("/grave        /acute      /circumflex      /tilde\n");
	printf("/dieresis      /hungarumlaut /ring      /caron\n");
	printf("/breve      /macron      /dotaccent      /cedilla\n");
	printf("/ogonek      /period      /grave      /acute\n");
	printf("/breve      /macron      /hyphen      /slash\n");
	printf("/.notdef      /.notdef      /.notdef      /.notdef\n");
	printf("/.notdef      /dotlessi      /.notdef      /.notdef\n");
	printf("/.notdef      /.notdef      /.notdef      /.notdef\n");
	printf("/space        /exclam       /quotedbl     /numbersign\n");
	printf("/dollar       /percent      /ampersand    /quoteright\n");
	printf("/parenleft    /parenright   /asterisk     /plus\n");
	printf("/comma        /hyphen       /period       /slash\n");
	printf("/zero         /one          /two          /three\n");
	printf("/four         /five         /six          /seven\n");
	printf("/eight        /nine         /colon        /semicolon\n");
	printf("/less         /equal        /greater      /question\n");
	printf("/at           /A            /B            /C\n");
	printf("/D            /E            /F            /G\n");
	printf("/H            /I            /J            /K\n");
	printf("/L            /M            /N            /O\n");
	printf("/P            /Q            /R            /S\n");
	printf("/T            /U            /V            /W\n");
	printf("/X            /Y            /Z            /bracketleft\n");
	printf("/backslash    /bracketright /asciicircum  /underscore\n");
	printf("/quoteleft    /a            /b            /c\n");
	printf("/d            /e            /f            /g\n");
	printf("/h            /i            /j            /k\n");
	printf("/l            /m            /n            /o\n");
	printf("/p            /q            /r            /s\n");
	printf("/t            /u            /v            /w\n");
	printf("/x            /y            /z            /braceleft\n");
	printf("/bar          /braceright   /tilde        /tilde\n");

/* Extensions to Latin-1 encoding: */

/* 0x80, 0x81, 0x82, 0x83 == octal 200, 201, 202, 203 */
	printf("/endash      /emdash        /quotedblleft  /quotedblright\n");
/* 0x84, 0x85, 0x86, 0x87 == octal 204, 205, 206, 207 */
	printf("/fi          /fl            /ellipsis     /.notdef\n");
/* 0x88, 0x89, 0x8a, 0x8b == octal 210, 211, 213, 214 */
	printf("/.notdef       /.notdef    /.notdef        /.notdef\n");
/* 0x8c, 0x8d, 0x8e, 0x8f == octal 214, 215, 216, 217 */
	printf("/.notdef   /.notdef        /.notdef         /.notdef\n");
/* 0x90, 0x91, 0x92, 0x93 == octal 220, 221, 223, 224 */
	printf("/periodcentered /lslash      /Lslash      /scaron\n");
/* 0x94, 0x95, 0x96, 0x97 == octal 224, 225, 226, 227 */
	printf("/Scaron       /zcaron       /Zcaron       /quotesingle\n");
/* 0x98, 0x99, 0x9a, 0x9b == octal 230, 231, 233, 234 */
	printf("/oe           /OE      /dagger      /daggerdbl\n");
/* 0x9c, 0x9d, 0x9e, 0x9f == octal 234, 235, 236, 237 */
	printf("/florin      /perthousand      /.notdef      /.notdef\n");

	printf("/space        /exclamdown   /cent         /sterling\n");
	printf("/currency     /yen          /brokenbar    /section\n");
	printf("/dieresis     /copyright    /ordfeminine  /guillemotleft\n");
	printf("/logicalnot   /hyphen       /registered   /macron\n");
	printf("/degree       /plusminus    /twosuperior  /threesuperior\n");
	printf("/acute        /mu           /paragraph    /bullet\n");
	printf("/cedilla      /onesuperior  /ordmasculine /guillemotright\n");
	printf("/onequarter   /onehalf      /threequarters /questiondown\n");
	printf("/Agrave       /Aacute       /Acircumflex  /Atilde\n");
	printf("/Adieresis    /Aring        /AE           /Ccedilla\n");
	printf("/Egrave       /Eacute       /Ecircumflex  /Edieresis\n");
	printf("/Igrave       /Iacute       /Icircumflex  /Idieresis\n");
	printf("/Eth          /Ntilde       /Ograve       /Oacute\n");
	printf("/Ocircumflex  /Otilde       /Odieresis    /multiply\n");
	printf("/Oslash       /Ugrave       /Uacute       /Ucircumflex\n");
	printf("/Udieresis    /Yacute       /Thorn        /germandbls\n");
	printf("/agrave       /aacute       /acircumflex  /atilde\n");
	printf("/adieresis    /aring        /ae           /ccedilla\n");
	printf("/egrave       /eacute       /ecircumflex  /edieresis\n");
	printf("/igrave       /iacute       /icircumflex  /idieresis\n");
	printf("/eth          /ntilde       /ograve       /oacute\n");
	printf("/ocircumflex  /otilde       /odieresis    /divide\n");
	printf("/oslash       /ugrave       /uacute       /ucircumflex\n");
	printf("/udieresis    /yacute       /thorn        /ydieresis\n");
	printf("] def\n");
	printf("%%%%EndResource\n");
}

void setfontnames() {

    switch (bfont) {
	default:
	case 0:
		fontname[ROMAN] = "Times-Roman";
		slanty_width =
		roman_width = tir_width;
		fontname[ITALIC] = "Times-Italic";
		italic_width = tii_width;
		fontname[BOLD] = "Times-Bold";
		bold_width = tib_width;
		fontname[BITALIC] = "Times-BoldItalic";
		bitalic_width = tibi_width;
		break;
	case 1:
		fontname[ROMAN] = "Palatino-Roman";
		slanty_width =
		roman_width = por_width;
		fontname[ITALIC] = "Palatino-Italic";
		italic_width = poi_width;
		fontname[BOLD] = "Palatino-Bold";
		bold_width = pob_width;
		fontname[BITALIC] = "Palatino-BoldItalic";
		bitalic_width = pobi_width;
		break;
	case 2:
		fontname[ROMAN] = "Bookman-Light";
		slanty_width =
		roman_width = bkl_width;
		fontname[ITALIC] = "Bookman-LightItalic";
		italic_width = bkli_width;
		fontname[BOLD] = "Bookman-Demi";
		bold_width = bkd_width;
		fontname[BITALIC] = "Bookman-DemiItalic";
		bitalic_width = bkdi_width;
		break;
	case 3:
		fontname[ROMAN] = "NewCenturySchlbk-Roman";
		slanty_width =
		roman_width = ncr_width;
		fontname[ITALIC] = "NewCenturySchlbk-Italic";
		italic_width = nci_width;
		fontname[BOLD] = "NewCenturySchlbk-Bold";
		bold_width = ncb_width;
		fontname[BITALIC] = "NewCenturySchlbk-BoldItalic";
		bitalic_width = ncbi_width;
		break;
	case 4:
		fontname[ROMAN] = "AvantGarde-Book";
		slanty_width =
		roman_width = agw_width;
		fontname[ITALIC] = "AvantGarde-BookOblique";
		italic_width = agwo_width;
		fontname[BOLD] = "AvantGarde-Demi";
		bold_width = agd_width;
		fontname[BITALIC] = "AvantGarde-DemiOblique";
		bitalic_width = agdo_width;
		break;
    }
	fontusage[ROMAN] = 0;
	fontusage[ITALIC] = 0;
	fontusage[BOLD] = 0;
	fontusage[BITALIC] = 0;

	fontname[SANS] = "Helvetica";
	fontusage[SANS] = 0;
	sans_width = hv_width;

	fontusage[SLANTY] = 0;

	fontname[MATH] = "Symbol";
	fontusage[MATH] = 0;

	fontname[TTYPE] = "Courier";
	fontusage[TTYPE] = 0;
	/* ttype_width = courier_width;*/

	fontname[ZD] = "ZapfDingbats";
	fontusage[ZD] = 0;

	fontname[IPA] = "xipa10";
	fontusage[IPA] = 0;

#ifdef USEMUSIX
	fontname[MUSIX] = "TeXMUSIX13-Regular";
	fontusage[MUSIX] = 0;
#endif

	default_width = courier_width;
}

static void setup() {

	printf("%%%%BeginSetup\n");
	printf("%%%%IncludeResource: font %s\n", fontname[ROMAN]);
	printf("%%%%IncludeResource: font %s\n", fontname[SANS]);
	printf("%%%%IncludeResource: font %s\n", fontname[ITALIC]);
	printf("%%%%IncludeResource: font %s\n", fontname[BOLD]);
	printf("%%%%IncludeResource: font %s\n", fontname[TTYPE]);
	printf("%%%%IncludeResource: font %s\n", fontname[IPA]);
#ifdef USEMUSIX
	printf("%%%%IncludeResource: font %s\n", fontname[MUSIX]);
#endif
	printf("%%%%IncludeResource: font %s\n", fontname[BITALIC]);
	printf("%%%%IncludeResource: font %s\n", fontname[MATH]);
	printf("%%%%IncludeResource: font %s\n", fontname[ZD]);

	printf("%% Fonts.\n");
	printf("/%s /FM-Roman MF\n", fontname[ROMAN]);
	printf("/RM { /FM-Roman %s %s 0 SF } def\n", ptsize, ptsize);
	printf("/SU { /FM-Roman %s %s 0 SF } def\n", superptsize, superptsize);
	printf("/TITLE { /FM-Roman %s %s 0 SF } def\n", titleptsize, titleptsize);
	printf("/%s /FM-Sans MF\n", fontname[SANS]);
	printf("/SA { /FM-Sans %s %s 0 SF } def\n", ptsize, ptsize);
	printf("/%s /FM-Italic MF\n", fontname[ITALIC]);
	printf("/IT { /FM-Italic %s %s 0 SF } def\n", ptsize, ptsize);
	printf("/SL { /FM-Roman %s %s %s SF } def\n", ptsize, ptsize, it_angle);
	printf("/%s /FM-Bold MF\n", fontname[BOLD]);
	printf("/BF { /FM-Bold %s %s 0 SF } def\n", ptsize, ptsize);
	printf("/%s /FM-Ttype MF\n", fontname[TTYPE]);
	printf("/TT { /FM-Ttype %s %s 0 SF } def\n", condensedsize, ttypeptsize);
	printf("/%s /FM-Bitalic MF\n", fontname[BITALIC]);
	printf("/BI { /FM-Bitalic %s %s 0 SF } def\n", ptsize, ptsize);
	printf("\n");


	printf("/MATH { /%s findfont %s scalefont setfont } def\n", fontname[MATH], ptsize);
	printf("/ZD { /%s findfont %s scalefont setfont } def\n", fontname[ZD], ptsize);
	printf("/IPA { /%s findfont %s scalefont setfont } def\n", fontname[IPA], ptsize);
	printf("/IPS { /%s findfont %s scalefont setfont } def\n",
		 fontname[IPA], superptsize);
#ifdef USEMUSIX
	printf("/MU { /%s findfont %s scalefont setfont } def\n", fontname[MUSIX], ptsize);
#endif
	printf("\n");


	printf("/#copies %d def\n", numcopies);

	printf("%% Pagedevice definitions:\n");
	printf("gs_languagelevel 1 gt {\n");
	printf("  <<\n");
	printf("    /Collate true /PageSize [612 792] \n");
	printf("  >> setpagedevice\n");
	printf("} if\n");
	printf("%%%%EndSetup\n");
}


void pagesetup(int pn) {
	int w;

	for (w = 0; w < MAXTL; w++)
		treeline[w].hpos1 = treeline[w].hpos2 = -1;

	ps_page = pn;

	printf("%%%%Page: (%d) %d\n", pn, pn);
	printf("%%%%BeginPageSetup\n");
	printf("_S\n");
	printf("%d %d translate\n", LEFT_RESERVE, TOP_RESERVE);
	printf("/pagenum %d def\n", pn);

	curr_a = 0;
	setfont(ROMAN);
	printf("%.1f setlinewidth\n", (float)ps_pointsize10/150.0);

	printf("%%%%EndPageSetup\n");

	if (!nopagenumbers && pageattop) {
		w = strlen(pageno_str);
		w = pwidth(pageno_str, ROMAN, w);
		printf("%d %d M",
	    	72 * ( (hsize_sp - w)/2 + offset_sp ) / ps_pitch_sp,
	    	792 - 72 - TOP_RESERVE + 24 - ps_voffset);
		printf("  (%s) s\n", pageno_str);
	}
}

void pagefinish() {
	int w;
	int text_size = vsize * ps_baselineskip;
	int pn_place;

	pn_place = 792 - text_size - 72 - TOP_RESERVE - 24 - ps_voffset;
	if (pn_place < ps_baselineskip/2) pn_place = ps_baselineskip/2;
 
	if (!nopagenumbers && !pageattop) {
		w = strlen(pageno_str);
		w = pwidth(pageno_str, ROMAN, w);
		printf("%d %d M",
	    	72 * ( (hsize_sp - w)/2 + offset_sp ) / ps_pitch_sp,
		pn_place);
		printf("  (%s) s\n", pageno_str);
	}

	printf("_R\nS\n");
}

static void trailer(int total) {
	int i = 0;

	printf("%%%%Trailer\n");
	printf("%%%%Pages: %d\n", total);

	printf("%%%%DocumentNeededResources: font");

	if (fontusage[ROMAN] || fontusage[SLANTY]) { printf(" %s", fontname[ROMAN]); i++; }
	if (fontusage[ITALIC]) { printf(" %s", fontname[ITALIC]); i++; }
	if (fontusage[BOLD]) {
	  if (i > 1) { printf("\n%%%%+ font"); i = 0; }
	  printf(" %s", fontname[BOLD]); i++; }
	if (fontusage[BITALIC]) {
	  if (i > 1) { printf("\n%%%%+ font"); i = 0; }
	  printf(" %s", fontname[BITALIC]); i++; }
	if (fontusage[TTYPE]) {
	  if (i > 1) { printf("\n%%%%+ font"); i = 0; }
	  printf(" %s", fontname[TTYPE]); i++; }
	if (fontusage[IPA]) {
	  if (i > 1) { printf("\n%%%%+ font"); i = 0; }
	  printf(" %s", fontname[IPA]); i++; }
#ifdef USEMUSIX
	if (fontusage[MUSIX]) {
	  if (i > 1) { printf("\n%%%%+ font"); i = 0; }
	  printf(" %s", fontname[MUSIX]); i++; }
#endif
	if (fontusage[MATH]) {
	  if (i > 1) { printf("\n%%%%+ font"); i = 0; }
	  printf(" %s", fontname[MATH]); i++; }
	if (fontusage[ZD]) {
	  if (i > 1) { printf("\n%%%%+ font"); i = 0; }
	  printf(" %s", fontname[ZD]); i++; }
	if (fontusage[SANS]) {
	  if (i > 1) { printf("\n%%%%+ font"); i = 0; }
	  printf(" %s", fontname[SANS]); i++; }
	printf("\n");

	printf("%%%%EOF\n");
}

static void setfont(unsigned char a) {
    if (curr_a == a) return;
    fontusage[a]++;
    switch (a) {
	case   BOLD: printf("BF\n"); break;
	case ITALIC: printf("IT\n"); break;
	case   MATH: printf("MATH\n"); break;
	case     ZD: printf("ZD\n"); break;
	case  TTYPE: printf("TT\n"); break;
	case    IPA: printf("IPA\n"); break;
#ifdef USEMUSIX
	case  MUSIX: printf("MU\n"); break;
#endif
	case SLANTY: printf("SL\n"); break;
	case  ROMAN: printf("RM\n"); break;
	case BITALIC: printf("BI\n"); break;
	case  SANS: printf("SA\n"); break;
	case SUPER:
	case SUB: printf("SU\n"); break;
	case IPASUPER:
	case IPASUB: printf("IPS\n"); break;
	case TITLE: printf("TITLE\n"); break;
	default: printf(" %% IMPOSSIBLE FONT %d\n",a); return;
    }
    curr_a = a;
}

#define OVERCLEARANCE 75
#define UNDERCLEARANCE 66

static int kern_ipa(unsigned char a, unsigned char b,
		 int *v, int *sk, int *ns, unsigned char at2) {
	int i;
	*v = *sk = *ns = 0;
	if (!ipa_width[a]) return 0;
	if (!ipa_width[b]) {
		int up = 0;
		int twa = ipa_true_width[a];
		int twb = ipa_true_width[b];
		int roomabove = OVERCLEARANCE;
		int roombelow = UNDERCLEARANCE;


		if (at2 == IPAOVER) up = 1;
		else if (at2 == IPAUNDER) up = -1;
		else if (b == 14) up = -1; /* subscript grave */
		else if (b == 15) up = -1; /* subscript acute */
		else if (b == 16) up = -1; /* subscript arch */
		else if (b == 34) up = -1; /* syllabicity */
		else if (b == 60) up = -1; /* bottom tie bar */
		else if (b == 38) up = 0; /* superimposed tilde (move down) */
		else if (ipa_y2[b] > -15 && ipa_y2[b] < 20) up = 0;
		else if (ipa_y2[b] < 300) up = -1;
		else up = 1;
/*
	|---aa---| |-bb-|
	          ^
	|---aa---| |-bb-|
	   ^....        
	|---aa---| |-bb-|
	          ^     
	|-aa-| |---bb---|
	      ^    
	|-aa-| |---bb---|
	 ^...   .....              
	|-aa-| |---bb---|
	         ^      
*/
		if (b == 60 || b == 62) {/* tie bars */
			/* this goes back too far */
			*sk = - twb/2;
			*ns = -(twb + *sk);
		}
		else {
			*sk = - (twa + twb)/2;
			*ns = twa + *sk;
		}

		if (b == 11 || b == 35 || b == 36 || b == 37 || b == 131 || b == 132 || b == 133) roombelow = 0;

		if (b == 38) *v = ipa_y1[a]/2 - ipa_y1[38]/2 - 50;
		else if (up > 0) *v = roomabove - (ipa_y1[b] - ipa_y2[a]); /* was 58 */
		else if (up < 0) *v = (ipa_y1[a] - ipa_y2[b]) - roombelow; /* was 58 */
		return 0;
	}
	if (at2 == IPAUNDER) *v = -250;
	else if (at2 == IPAOVER) *v = 250;

	for (i = 0; ipa_kern[i][0] ; i++) {
		if (ipa_kern[i][0] == a && ipa_kern[i][1] == b)
			return ipa_kern[i][2];
	}
	return 0;
}

#define KPROC(F) \
static int kern_ ## F(unsigned char a, unsigned char b, \
		 int *v, int *sk, int *ns, unsigned char at2) { \
	int i; \
	if (roman_diacritic[a] == TRUE) return 0; \
	if (roman_diacritic[b] == TRUE) { \
		int twa = F ## _width[a]; \
		int twb = F ## _width[b]; \
		*sk = - (twa + twb)/2; \
		*ns = twa + *sk; \
		if (b==PS_cedilla||b==PS_ogonek) \
			*v = (F ## _y1[a] - F ## _y2[b]); \
		else if (b==PS_underdot||b==PS_underline||b==PS_undergrave||b==PS_underacute||b==PS_underbreve) \
			*v = (F ## _y1[a] - F ## _y2[b]) - UNDERCLEARANCE; \
		else if (b==PS_stroke||b==PS_slash) *v = 0; \
		else *v = OVERCLEARANCE - (F ## _y1[b] - F ## _y2[a]); \
		return 0; \
	} \
	for (i = 0; F ## _kern[i][0] ; i++) { \
		if (F ## _kern[i][0] == a && F ## _kern[i][1] == b) \
			return F ## _kern[i][2]; \
	} \
	return 0; \
}

KPROC(hv)
KPROC(tir)
KPROC(tii)
KPROC(tib)
KPROC(tibi)

KPROC(por)
KPROC(poi)
KPROC(pob)
KPROC(pobi)

KPROC(ncr)
KPROC(nci)
KPROC(ncb)
KPROC(ncbi)

KPROC(bkl)
KPROC(bkli)
KPROC(bkd)
KPROC(bkdi)

KPROC(agw)
KPROC(agwo)
KPROC(agd)
KPROC(agdo)

KPROC(ttype)

static int kern(unsigned char at, unsigned char a, unsigned char b,
	 int *v, int *sk, int *ns, unsigned char at2) {
	int ret = 0;
	if (!is_ipa(at)) {
		*v = *sk = *ns = 0;
		if ( (a=isoprint[a]) == DEL ) return 0;
		if (b != PS_tilde) if ( (b=isoprint[b]) == DEL ) return 0;
	}
	switch (at) {
	    case SLANTY:
	    case ROMAN: switch (bfont) {
	    		case 0: return kern_tir(a,b,v,sk,ns,at2);
			case 1: return kern_por(a,b,v,sk,ns,at2);
			case 2: return kern_bkl(a,b,v,sk,ns,at2);
			case 3: return kern_ncr(a,b,v,sk,ns,at2);
			case 4: return kern_agw(a,b,v,sk,ns,at2);
			}
			break;
	    case SUPER:
	    case SUB: switch (bfont) {
	    		case 0: ret = 3 * kern_tir(a,b,v,sk,ns,at2) / 4;
			case 1: ret = 3 * kern_por(a,b,v,sk,ns,at2) / 4;
			case 2: ret = 3 * kern_bkl(a,b,v,sk,ns,at2) / 4;
			case 3: ret = 3 * kern_ncr(a,b,v,sk,ns,at2) / 4;
			case 4: ret = 3 * kern_agw(a,b,v,sk,ns,at2) / 4;
			}
			*v = 3 * *v / 4;
			*sk = 3 * *sk / 4;
			*ns = 3 * *ns / 4;
		        return ret;
			break;
	    case TITLE: switch (bfont) {
	    		case 0: ret = 2 * kern_tir(a,b,v,sk,ns,at2);
			case 1: ret = 2 * kern_por(a,b,v,sk,ns,at2);
			case 2: ret = 2 * kern_bkl(a,b,v,sk,ns,at2);
			case 3: ret = 2 * kern_ncr(a,b,v,sk,ns,at2);
			case 4: ret = 2 * kern_agw(a,b,v,sk,ns,at2);
			}
			*v = 2 * *v;
			*sk = 2 * *sk;
			*ns = 2 * *ns;
		        return ret;
			break;
	    case ITALIC: switch (bfont) {
	    		case 0: return kern_tii(a,b,v,sk,ns,at2);
			case 1: return kern_poi(a,b,v,sk,ns,at2);
			case 2: return kern_bkli(a,b,v,sk,ns,at2);
			case 3: return kern_nci(a,b,v,sk,ns,at2);
			case 4: return kern_agwo(a,b,v,sk,ns,at2);
			}
			break;
	    case BOLD:  switch (bfont) {
	    		case 0: return kern_tib(a,b,v,sk,ns,at2);
			case 1: return kern_pob(a,b,v,sk,ns,at2);
			case 2: return kern_bkd(a,b,v,sk,ns,at2);
			case 3: return kern_ncb(a,b,v,sk,ns,at2);
			case 4: return kern_agd(a,b,v,sk,ns,at2);
			}
			break;
	    case BITALIC: switch (bfont) {
	    		case 0: return kern_tibi(a,b,v,sk,ns,at2);
			case 1: return kern_pobi(a,b,v,sk,ns,at2);
			case 2: return kern_bkdi(a,b,v,sk,ns,at2);
			case 3: return kern_ncbi(a,b,v,sk,ns,at2);
			case 4: return kern_agdo(a,b,v,sk,ns,at2);
			}
			break;
	    case TTYPE:  return kern_ttype(a,b,v,sk,ns,at2);
	    case SANS:  return kern_hv(a,b,v,sk,ns,at2);
	    case IPASUPER:
	    case IPASUB:   ret = 3 * kern_ipa(a,b,v,sk,ns,at2) / 4;
			   *v = 3 * *v / 4;
			   *sk = 3 * *sk / 4;
			   *ns = 3 * *ns / 4;
			   return ret;
	    case IPAOVER:
	    case IPAUNDER:
	    case IPA:   return kern_ipa(a,b,v,sk,ns,at2);
	    case MUSIX:
	    default:     return 0;
	}
	return 0;
}
static int does_kern(unsigned char at) {
	switch (at) {
	    case TITLE:
	    case SLANTY:
	    case ROMAN: 
	    case SUPER: 
	    case SUB: 
	    case ITALIC:
	    case BOLD:
	    case BITALIC:
	    case IPA:
	    case IPASUPER:
	    case IPASUB:
	    case TTYPE:
	    case SANS:   return TRUE;
	    case IPAOVER:
	    case IPAUNDER:
	    default:     return FALSE;
	}
}



int pnwidth(unsigned char *s, unsigned char *a, int n) {
	unsigned char *p, *q, *lw, at, last_a;
	int w = 0, sw = 0, ch, lwlen = 0;

	last_a = ROMAN;
	lw = s;

	for (p = s, q = a ; n > 0; n--, p++, q++, lwlen++) {
		if (*p == ' ') {
			at = last_a;
			lw = p;
			lwlen = 0;
		}
		else last_a = at = *q;
		ch = (int)*p;
		if (ch == LIT_ESC) {
			n--; p++; q++;
			if (n <= 0) {
				if (psflag) return w;
				return sw * 500;
			}
			ch = (int)*p;
		}
		else if (ch == COL_ESC) {
			n--; p++; q++;
			w = SP((*p + 1));
			sw = *p + 1;
			continue;
		}
		else if (ch == UNI_ESC) {
			unsigned int wc;
			n--; p++; q++;
			n--; p++; q++;
			if (n <= 0) {
				if (psflag) return w;
				return sw * 500;
			}
	    		wc = (unsigned int)((*p << 8) | *(p-1));
			if (wc >= 0x100) {
			    if (!is_combining(wc)) {
				sw++;
				w += SP(1);
			    }
			    continue;
			}
		}
		else if (ch == KERN_ESC) {
			int arg = *(p+1) - 1;
			int arg2 = *(p+2) - 1;
			int sp_sp;
			n--; p++; q++;
			n--; p++; q++;
			sp_sp = (arg * 100 + arg2 - 10000) * ps_pitch_sp / 72;
			w += sp_sp;
			sw += COL(sp_sp);
		}
		else if (ch == PS_ESC) {
			while (n > 0 && *p != ' '-1) { n--; p++; q++; }
		}
		if (!psflag) {
			if (!is_ipa(at) || ipa_width[ch]) sw++;
			continue;
		}
		if (is_roman(at) && roman_diacritic[ch] == TRUE) continue;
		switch (at) {
		    case SUPER:
		    case SUB: w += 3 * roman_width[ch] / 4; break;
		    case TITLE: w += 2 * roman_width[ch]; break;
		    case ROMAN: w += roman_width[ch]; break;
		    case ITALIC: w += italic_width[ch]; break;
		    case BOLD: w += bold_width[ch]; break;
		    case BITALIC: w += bitalic_width[ch]; break;
		    case SLANTY: w += slanty_width[ch]; break;
		    case MATH: w += sym_width[ch]; break;
		    case ZD: w += zd_width[ch]; break;
		    case TTYPE: w += ttype_width[ch]; break;
		    case SANS: w += sans_width[ch]; break;
		    case IPAOVER:
		    case IPAUNDER:
		    case IPA: w += ipa_width[ch]; break;
		    case IPASUPER:
		    case IPASUB: w += 3 * ipa_width[ch] / 4; break;
		    case MUSIX: w += musix_width[ch]; break;
		    case BOX: w += SP(1); break;
		    default: fprintf(stderr,"at=%d\n",at); w += default_width[ch]; break;
		}
	}
	if (psflag) return w;
	return sw * 500;
}

int pwidth(unsigned char *s, unsigned char at, int n) {
	unsigned char *p;
	int w = 0, sw = 0, ch;
	for (p = s; n > 0; n--,p++) {
		ch = (int)*p;
		if (ch == LIT_ESC) {
			n--; p++;
			if (n <= 0) {
				if (psflag) return w;
				return sw;
			}
			ch = (int)*p;
		}
		else if (ch == COL_ESC) {
			n--; p++;
			if (n <= 0) {
				if (psflag) return w;
				return sw;
			}
			w = SP((*p + 1));
			sw = *p + 1;
			continue;
		}
		else if (ch == UNI_ESC) {
			unsigned int wc;
			n--; p++;
			n--; p++;
			if (n <= 0) {
				if (psflag) return w;
				return sw;
			}
	    		wc = (unsigned int)((*p << 8) | *(p-1));
			if (wc >= 0x100) {
			    if (!is_combining(wc)) {
				sw++;
				w += SP(1);
			    }
			    continue;
			}
		}
		else if (ch == KERN_ESC) {
			int arg = *(p+1) - 1;
			int arg2 = *(p+2) - 1;
			int sp_sp;
			n--; p++;
			n--; p++;
			sp_sp = (arg * 100 + arg2 - 10000) * ps_pitch_sp / 72;
			w += sp_sp;
			sw += COL(sp_sp);
		}
		else if (ch == PS_ESC) {
			while (n > 0 && *p != ' '-1) { n--; p++; }
		}
		if (!psflag) {
			if (!is_ipa(at) || ipa_width[ch]) sw++;
			continue;
		}
		if (is_roman(at) && roman_diacritic[ch] == TRUE) continue;
		switch (at) {
		    case SUPER:
		    case SUB: w += 3 * roman_width[ch] / 4; break;
		    case TITLE: w += 2 * roman_width[ch]; break;
		    case ROMAN: w += roman_width[ch]; break;
		    case ITALIC: w += italic_width[ch]; break;
		    case BOLD: w += bold_width[ch]; break;
		    case BITALIC: w += bitalic_width[ch]; break;
		    case SLANTY: w += slanty_width[ch]; break;
		    case MATH: w += sym_width[ch]; break;
		    case ZD: w += zd_width[ch]; break;
		    case TTYPE: w += ttype_width[ch]; break;
		    case SANS: w += sans_width[ch]; break;
		    case IPAOVER:
		    case IPAUNDER:
		    case IPA: w += ipa_width[ch]; break;
		    case IPASUPER:
		    case IPASUB: w += 3 * ipa_width[ch] / 4; break;
		    case MUSIX: w += musix_width[ch]; break;
		    case BOX: w += SP(1); break;
		    default: w += default_width[ch]; break;
		}
	}
	if (psflag) return w;
	return sw;
}

static void makerule(int what, int deltax) {
	float x;
    	float bottom = (float)ps_baselineskip10/20.0;
	if (what == ' ') return;

	x = 72.0 * deltax / ps_pitch_sp;
	printf(" 0 %.2f rmoveto\n", bottom/2.0);
	printf("  %.1f  0  rlineto ", x);
	if (what == '.') printf("[1 3] 0 setdash stroke [] 0 setdash\n");
	else printf("stroke\n");
}

static void makeboxes(unsigned char *s, int len, int x, int y) {
    float center = 72.0 * 250 / ps_pitch_sp;
    float bottom = (float)ps_baselineskip10/20.0;
    int b, tlen = 0;
    int bold;

    while (len) {
	len--; tlen++;
	b = *s++;
	bold = 0;
	if (isupper(b)) {
		bold = 1;
		b = tolower(b);
	}
	printf(" 0 %.2f rmoveto\n", bottom/2.0);
	/* position for vertical line */
	switch(b) {
	    case 'x':
	    case 't':
	    case 'u':
	    case 'w':
	    case 'n':
	    case 'l':
	    case 'k':
		printf(" %.2f -%.2f rmoveto", center, bottom);
		break;
	    case 'm':
	    case 'j':
	    case 'v':
		printf(" %.2f 0 rmoveto", center);
		break;
	    case 'h':
		printf(" 0 -%.2f rmoveto", bottom);
		break;
	    case 'q':
	    default:
		break;
	}
	/* draw vertical line */
	switch(b) {
	    case 'x':
	    case 't':
	    case 'u':
	    case 'n':
	    case 'h':
		/* full length vertical */
		printf(" 0 %.2f rlineto", 2*bottom);
		break;
	    case 'v':
	    case 'w':
	    case 'l':
	    case 'k':
	    case 'm':
	    case 'j':
		/* half length vertical */
		printf(" 0 %.2f rlineto", bottom);
		break;
	    case 'q':
	    default:
		break;
	}
	/* position for horizontal line */
	switch(b) {
	    case 'n':
	    case 'u':
	    case 'v':
	    case 'j':
		printf(" -%.2f -%.2f rmoveto", center, bottom);
		break;
	    case 'm':
	    case 't':
		printf(" 0 -%.2f rmoveto", bottom);
		break;
	    case 'k':
	    case 'w':
		printf(" -%.2f 0 rmoveto", center);
		break;
	    case 'x':
	    case 'q':
	    case 'l':
	    case 'h':
	    default:
		break;
	}
	/* draw horizontal line */
	switch(b) {
	    case 't':
	    case 'u':
	    case 'l':
	    case 'k':
	    case 'm':
	    case 'j':
		/* half length horizontal */
		printf(" %.2f 0 rlineto", center);
		break;
	    case 'w':
	    case 'q':
	    case 'n':
	    case 'h':
	    case 'v':
		/* full length horizontal */
		printf(" %.2f 0 rlineto", 2*center);
		break;
	    case 'x':
	    default:
		break;
	}
	if (b == 'h') {
		/* full length down vertical */
		printf(" 0 -%.2f rlineto", 2*bottom);
		/* full length back horizontal */
		printf(" closepath");
	}

	if (bold) {
		if (b == 'h') printf(" .25 setgray fill 0 setgray\n");
		else printf(" currentlinewidth dup 2 mul setlinewidth stroke setlinewidth\n");
	}
	else {
		if (b == 'h') printf(" .50 setgray fill 0 setgray\n");
		else printf(" stroke\n");
	}

	if (len) move_to(x + 500 * tlen, y);
    }
}

/* test node for value of attribute */
static char has(int i, char attrib)
{
	if (attrib >= 'A' && attrib <= 'Z')
		return(treeline[i].attrib[attrib - 'A']);
	else return(0);
}


static void draw_line(int i) {
	int status, x1, y1, x2, y2, yupadjust, ydownadjust;
	float fx1, fx2;

	status = treeline[i].status;
	treeline[i].status = 0;

	printf(" %% line %d from #%d(%d,%d) to #%d(%d,%d)",
		i,
		treeline[i].id1,
		COL(treeline[i].hpos1), treeline[i].vpos1,
		treeline[i].id2,
		COL(treeline[i].hpos2), treeline[i].vpos2);
	if (status == 1 && has(i,'V')) printf(" is vertical\n");
	else printf("\n");

	if (status == 1 && has(i,'L')) {
		printf(" %% no line from lexical node\n");
		return;
	}
	if (status == 1 && has(i,'O') && (!has(i,'C') || has(i,'D'))) {
		printf(" %% line is omitted\n");
		return;
	}
	if (status == 1 && has(i,'R')) {
		printf(" %% no line from relational node\n");
		return;
	}
	x1 = treeline[i].hpos1;
	x2 = treeline[i].hpos2;

	if (status == 1 && has(i,'V')) {
/*
		x1 = (x1 + x2)/2;
		x2 = x1;
*/
/*
 * Not sure this is necessary anymore.
*/
		x1 = x2;
	}
	if (status == 3) { /* triangle base */
		if (treeline[i].left1) x1 = treeline[i].left1;
	}

	y1 = treeline[i].vpos1;
	y2 = treeline[i].vpos2;

	/* yupadjust = (3*ps_baselineskip10 + 5) / 40; */
	yupadjust = (3*ps_baselineskip10 + 5) / 35;
	/* ydownadjust = (ps_baselineskip10 + 5) / 20; */
	ydownadjust = (2*ps_baselineskip10 + 5) / 45;

    if (status == 4) {
	if (treeline[i].named1) y1 += yupadjust;
	if (treeline[i].named2) y2 -= ydownadjust;
    }
    else {
	if (y1 < y2) { /* pointing up */
		if (treeline[i].named1) y1 += yupadjust;
		if (treeline[i].empty2) y2 += yupadjust;
		else if (treeline[i].named2) y2 -= ydownadjust;
	}
	else if (y1 > y2) {
		if (treeline[i].empty1) y1 += yupadjust;
		else if (treeline[i].named1) y1 -= ydownadjust;
		if (treeline[i].named2) y2 += yupadjust;
	}
	else {
	/* base of triangle */
		if (treeline[i].named1) y1 += yupadjust;
		if (treeline[i].named2) y2 += yupadjust;
		x2 = treeline[i].right2;
	}
    }

	fx1 = 72.0 * x1 / ps_pitch_sp;
	fx2 = 72.0 * x2 / ps_pitch_sp;

	if (status == 4) {
		int deltay;
		printf("%.1f %d M ", fx1, y1);
		/* Daughter is at (fx1,y1) and mother at (fx2,y2);
		 */
		if (y1 < y2) deltay = y2 - y1;
		else deltay = y1 - y2;
		if (deltay < ps_baselineskip10/5) deltay = ps_baselineskip10/5;

		printf("%.1f %d ", fx1, y1 + deltay);
		printf("%.1f %d ", fx2, y2 - deltay);
		printf("%.1f %d curveto ", fx2, y2);
	}
	else {
		printf(" %.1f %d M %.1f %d lineto ", fx1, y1, fx2, y2);
	}
	if (has(i,'C')) printf("[1 3] 0 setdash stroke [] 0 setdash\n");
	else printf("stroke\n");

	if (status == 3) {
		int xt;
		float fxt;
		xt = treeline[i].right1;
		fxt = 72.0 * xt / ps_pitch_sp;
		if (xt) {
			printf(" %.1f %d M %.1f %d lineto ", fx1, y1, fxt, y1);
			if (has(i,'C')) printf("[1 3] 0 setdash stroke [] 0 setdash\n");
			else printf("stroke\n");
			printf(" %.1f %d M %.1f %d lineto ", fxt, y1, fx2, y2);
			if (has(i,'C')) printf("[1 3] 0 setdash stroke [] 0 setdash\n");
			else printf("stroke\n");
		}
	}
}

static void node_position(int id, int end1, int end2, int x, int y) {
	int i;
	for (i = 0; i < MAXTL; i++)
	   if (treeline[i].status >= 1 && treeline[i].id1 == id) {
		treeline[i].hpos1 = x;
		treeline[i].vpos1 = y;
		treeline[i].left1 = end1;
		treeline[i].right1 = end2;
		if (end1 && has(i,'S') == 'f') treeline[i].hpos1 = end1;
		if (end2 && has(i,'S') == 'l') treeline[i].hpos1 = end2;
		if (treeline[i].hpos2 >= 0) draw_line(i);
	  }

	for (i = 0; i < MAXTL; i++)
	   if (treeline[i].status >= 1 && treeline[i].id2 == id) {
		treeline[i].hpos2 = x;
		treeline[i].vpos2 = y;
		treeline[i].left2 = end1;
		treeline[i].right2 = end2;
		if (treeline[i].hpos1 >= 0) draw_line(i);
	   }
}

static void move_to(int this_x, int this_y) {
	char str_x[50];
	int w;
	float x;

	x = 72.0 * (float)this_x / (float)ps_pitch_sp;
	sprintf(str_x, "%.1f", x);
	w = strlen(str_x);
	if (str_x[w - 1] == '0') str_x[w - 2] = '\0';
	printf("%s %d M ", str_x, this_y);
}

#define MAXSEG 1000
static struct {
	unsigned char *s, *a, at;
	int col, width, nwidth, len, mid, treelen, type, aligned, lbearing, kerns, vkerns, wglue;
} ol[MAXSEG];

#define TEXT_SEG 5
#define CTL_SEG 0
#define HRULE_SEG 6

void ps_outline(int pline, int wh, int i)
{
	unsigned char *p, *q, a, last_a, t;
	int curr_tb, text_top, left_indent = 0, left_space = 0;
	int curr_y, glue = 0, total_width = 0;
	int seg, end_seg, col, last_aligned = -1;
	int this_x, len, col_indent;
	int curr_x, more_glue, curr_glue;
	int target, super = 0, sub = 0;
	int command, arg, arg2, arg3;

	curr_tb = hbox[wh][i].tb;
	len = hbox[wh][i].len;
	dotab(-1, curr_tb);

	/* need to do justification? */
	target = hbox[wh][i].tlen;

	text_top = Letter_Height - 72 - TOP_RESERVE - ps_voffset;
	curr_x = hbox[wh][i].spaces + offset_sp;
	if (curr_x < 0) curr_x = 0;

	/*col_indent = COL(hbox[wh][i].spaces);*/
	col_indent = hbox[wh][i].spaces;

	/* curr_y = text_top - pline * ps_baselineskip; */
	curr_y = text_top - pline;
	if (curr_y < ps_baselineskip) return;

	p = hbox[wh][i].ch;
	q = hbox[wh][i].at;

	#ifdef LOOKATSOURCE
		fprintf(stderr,"%d: ", pline);
		for (arg = 0; arg < len; arg++) {
			if (is_ctl(p[arg])) fprintf(stderr,"{%d}", p[arg]);
			else fprintf(stderr,"%c", p[arg]);
		}
		fprintf(stderr,"\n");
	#endif

	while (len && p[len-1] == ' ' && (len-1) && p[len-2] != LIT_ESC) len--;
	while (len && p[len-1] == '\t' && (len-1) && p[len-2] != LIT_ESC) len--;

	if (!len) return;

	while (left_space < len && p[left_space] == ' ') left_space++;
	if (left_space >= len) return;

	left_indent = pnwidth(p, q, left_space);
	p += left_space;
	q += left_space;
	len -= left_space;

	#ifdef TABDEBUG
		fprintf(stderr, "scan of line %d:", pline);
	#endif

	seg = 0;
	ol[0].s = p;
	ol[0].a = q;
	last_a = *q;
	col = 0;
	ol[0].col = 0;
	ol[0].len = 0;
	ol[0].treelen = 0;
	ol[0].lbearing = 0;
	ol[0].aligned = FALSE;
	ol[0].at = curr_a;
	t = ' ';

	for ( ; len > 0; len--, p++, q++, col++) {
		if (*p == '\t') {
			int tcol, pseg = seg;
			#ifdef TABDEBUG
				fprintf(stderr,"(tab[%d])",seg);
			#endif
			if (pseg) {
				pseg--;
				ol[pseg].aligned = TRUE;
				if (pseg > last_aligned) last_aligned = pseg;
			}
			glue = -1;
			tcol = dotab(col + col_indent, curr_tb) - col_indent;
			/*if (tcol - 1 > col)*/ col = tcol - 1;
			t = ' ';
			continue;
		}
	#ifdef TABDEBUG
		else
		fprintf(stderr,"{%c}",*p);
	#endif
		ol[seg].len++;
		if (*p <= ' ') {
			if (*p == ' ' && glue >= 0) glue++;
			t = *p;
			continue;
		}
		a = (*q & 0x7f);
		if (a == IPAOVER || a == IPAUNDER) a = IPA;
		if (t > ' ' && a == last_a && /*a != MATH && */
			 *p != HRULE_ESC && *p != PS_ESC &&
			 *p != NODE_ESC && *p != KERN_ESC &&
			 *p != COL_ESC) {
			t = *p;
			if (*p == LIT_ESC) {
				p++; q++; len--;
			}
			continue;
		}
	#ifdef TABDEBUG
		fprintf(stderr," col%d[%d]\n",col,seg);
	#endif
		ol[seg].at = a;
	#ifdef TABDEBUG
		if (a != last_a && (!a||a > BOX))
		printf("%% Er, last was %d, here is %d, p->%c, t=%c\n",last_a,a,*p,t);
	#endif
		/* last_a = a; */
		ol[seg].s = p;
		ol[seg].a = q;
		/*if (seg && ol[seg-1].type == HRULE_SEG) col--;*/ /*??*/
		ol[seg].col = col;
		ol[seg].width = ol[seg].nwidth = ol[seg].mid =
			 ol[seg].kerns = ol[seg].vkerns = 0;
		t = *p;
		if (*p == PS_ESC && *(p + 1) == '{') {
		/* scan over { ... } without advancing col */
			while (len > 0 && *p != ' '-1) { len--; p++; q++; }
			col--;
			t = ' ';
			ol[seg].type = CTL_SEG;
		}
		else if (*p == NODE_ESC) {
			int tcol, tlen, lbear, pseg = seg;
			tcol = *(p+1) - 1;
			tlen = *(p+2) - 1;
			lbear = *(p+3) - 1;
		/* Debug tree node positioning.
		printf("%%--at col=%d seg=%d, tcol=%d",
		 col, seg, tcol);
		*/
			ol[seg].type = CTL_SEG;
			ol[seg].lbearing = SP(lbear) / COLMUL;
			if (pseg) pseg--;
			if (pseg >= 0 && ol[pseg].type == TEXT_SEG) {
				ol[pseg].aligned = TRUE;
				if (pseg > last_aligned) last_aligned = pseg;
			}

			while (pseg > 0 && ol[pseg-1].type == TEXT_SEG )
				pseg--;
			tcol -= ol[pseg].col;
			if (tlen) {
				ol[pseg].treelen = 1 /*SP(tlen) / COLMUL */;
			}
		/*
		printf(" at pseg=%d, tcol diff=%d\n",
		 pseg, tcol);
		*/
			while (pseg <= seg) {
				if (!tlen) ol[pseg].treelen = -1;
				ol[pseg++].col += tcol;
			}
			col += tcol;
		/*
		printf("\nNODE_ESC, col=%d, arg=%d, arg2=%d, arg3=%d\n",
		 col, *(p+1), *(p+2), *(p+3) );
		*/
			p += 5;
			q += 5;
			len -= 5;
			col--;
			t = ' ';
			glue = -1;
		}
		else if (*p == COL_ESC) {
			int tcol, pseg = seg;
			p++; q++; len--;
			tcol = *p - 1;
			t = ' ';
			glue = -1;
			pseg--;
			if (pseg >= 0 && ol[pseg].type == TEXT_SEG) {
				ol[pseg].aligned = TRUE;
				if (pseg > last_aligned) last_aligned = pseg;
			}
			/*if (tcol - 1 > col + col_indent)*/
				 /* col = tcol - 1 - col_indent;*/
			col = tcol - 1;
			if (!seg) ol[seg].col = col;
			ol[seg].type = CTL_SEG;
			#ifdef TABDEBUG
			fprintf(stderr,"\nCOL_ESC tcol=%d col=%d\n",tcol,col);
			#endif
		}
		else if (*p == KERN_ESC) {
			p += 2;
			q += 2;
			len -= 2;
			col--;
			t = ' ';
			ol[seg].type = TEXT_SEG;
		}
		else if (*p == HRULE_ESC) {
			int tcol;
			p++; q++; len--;
			p++; q++; len--;
			tcol = *p << 8;
			p++; q++; len--;
			tcol += *p;
			tcol = COL(tcol);
			col += tcol - 1;
			t = ' ';
			last_a = a;
			ol[seg].type = HRULE_SEG;
		}
		else {
			last_a = a;
			ol[seg].type = TEXT_SEG;
			if (glue < 0 && !ol[seg].aligned &&
				seg > last_aligned) glue = 0;
		}
		if (*p == LIT_ESC) {
			p++; q++; len--;
		}
	#ifdef TABDEBUG
		fprintf(stderr,"END seg %d col = %d\n\n", seg, col);
	#endif
		seg++;
		ol[seg].aligned = FALSE;
		ol[seg].len = 0;
		ol[seg].treelen = 0;
		ol[seg].lbearing = 0;
	}
	ol[seg].s = p;
	ol[seg].a = q;
	ol[seg].col = col;
	ol[seg].type = CTL_SEG;
	end_seg = seg;

	for (seg = 0; seg < end_seg; seg++) {
		int pseg, bearing;
		ol[seg].len = ol[seg+1].s - ol[seg].s;
		ol[seg].mid = 0;
		p = ol[seg].s;
		if (ol[seg].type == CTL_SEG) {
			ol[seg].width = 0;
			ol[seg].wglue = 0;
			bearing = ol[seg].lbearing;
			if (bearing) {
				for (pseg = seg; seg > 0 && ol[pseg-1].type == TEXT_SEG; pseg--) ;
				if (!pseg) left_indent += bearing;
				else {
					pseg -= 2;
					if (pseg >= 0 && ol[pseg].type == TEXT_SEG && ol[pseg].aligned) {
						ol[pseg].width += bearing;
						printf("%% bearing %d added to seg %d\n", bearing, pseg);
					}
					else printf("%% lost bearing at seg %d\n", seg);
				}
				if (seg && ol[seg-1].type == TEXT_SEG && ol[seg-1].aligned) {
					ol[seg-1].width -= bearing;
				}
				else printf("%% stray bearing at seg %d\n", seg);
			}
			continue;
		}
		if (seg < end_seg - 1 &&
			 !frenchspacing &&
			 ol[seg].type == TEXT_SEG &&
			 !ol[seg].aligned &&
			 is_roman(ol[seg].at))
			ol[seg].wglue = endofsentence(p, ol[seg].len);
		else ol[seg].wglue = 0;
		if (ol[seg].len < 2 || *(p + ol[seg].len - 2) != ' ')
			ol[seg].wglue = 0;
		pseg = seg + 1;
		while (pseg < end_seg && ol[pseg].type == CTL_SEG) pseg++;

		ol[seg].width =
		  SP(ol[pseg].col) - SP(ol[seg].col);
	}

	left_indent += SP(ol[0].col);

	for (seg = 0; seg < end_seg; seg++) {
	    int vwidth, kcount, vkcount;
	    int pseg;
	    unsigned char at = ol[seg].at;

	    p = ol[seg].s;
	    q = ol[seg].a;
	    command = *p;
	    vwidth = ol[seg].width;
	    if (command == KERN_ESC) {
		arg = *(p+1) - 1;
		arg2 = *(p+2) - 1;
		printf("\n %% KERN=%d\n",
			 arg * 100 + arg2 - 10000);
		vwidth = (arg * 100 + arg2 - 10000) * ps_pitch_sp / 72;
/*
	ps_pitch_sp = 720000 / ps_pointsize10;
	x = 72.0 * (float)this_x / (float)ps_pitch_sp;
	x * ps_pitch_sp = 72 * this_x
	this_x = (x * ps_pitch_sp) / 72
say -p10, then ps_pitch_sp = 7200; say kern 3pt; this_x = 3 * 7200 / 72 = 300
72 * 300 / 7200 = 3
*/
	    }
	    else if (command == HRULE_ESC) {
		int tcol;
		p++;
		p++;
		tcol = *p << 8;
		p++;
		tcol += *p;
		vwidth = tcol;
	    }

	    if (ol[seg].type == TEXT_SEG && command != KERN_ESC) {

		/* now fix width for actual width of text */
		vwidth = pnwidth(ol[seg].s, ol[seg].a, ol[seg].len);
		#ifdef TABDEBUG
		fprintf(stderr,"vwidth[%d]=%d\n",seg,vwidth);
		#endif
		/* kcount is number of kerns to be printed */
		vkcount = kcount = 0;

		if (vwidth && does_kern(at)) {
		    unsigned char a, olda, b;
		    int litaflag = FALSE;
		    int litbflag = FALSE;
		    int k, v, sk, ns, jk;
		    a = ' ';
		    olda = 0;
		    for ( ; p < ol[seg+1].s; p++, q++) {
			b = *p;
			if (b == LIT_ESC && !litbflag) {
				litbflag = TRUE;
				continue;
			}
		/* add in kern between chars before and after diacritic */
			if ((a > ' ' || litaflag ) && (b > ' ' || litbflag)) {
				k = kern(at, a, b, &v, &sk, &ns, *q);
				if (!k && !sk && olda > 0) {
				    k = kern(at, olda, b, &jk, &jk, &jk, at);
				    olda = 0;
				}
				else if (!k && sk) olda = a;
				else olda = 0;
				if (k || sk) kcount++;
				if (v) vkcount++;
				vwidth += k;
			}
			/* kern with space at end of text changes width,
			 * but is not to be printed
			 */
			if (!is_ipa(at) && b == ' ' && !litbflag) {
				k = kern(at, a, b, &v, &sk, &ns, at);
				vwidth += k;
			}
			/* kern with space at beginning of text moves this text
			 * closer to preceding, but is not to be printed
			 */
			else if (!is_ipa(at) && a == ' ' && !litaflag) {
				k = kern(at, a, b, &v, &sk, &ns, at);
				if (seg && ol[seg-1].type == TEXT_SEG) {
					ol[seg-1].nwidth += k;
					if (!ol[seg-1].aligned)
					ol[seg-1].width += k;
				}
			}
			a = b;
			litaflag = litbflag;
			litbflag = FALSE;
		    }
		}
		ol[seg].kerns = kcount + vkcount;
		ol[seg].vkerns = vkcount;

	    } /* end if TEXT_SEG */

	    /* record the natural width */
	    ol[seg].nwidth = vwidth;

	    /* the midpoint will be used to draw tree lines */
	    if (ol[seg].treelen >= 0) ol[seg].mid = vwidth/2;

	    if (ol[seg].type == TEXT_SEG || ol[seg].type == HRULE_SEG) {

		/*if (ol[seg].wglue == 1) vwidth += COMMASPACE - 250;
		else*/ if (ol[seg].wglue > 1) vwidth += SENTENCESPACE - 250;

		/* possibly shrink the width down to the natural width */
		if (vwidth && (!ol[seg].aligned || seg + 1 == end_seg)) {
			for (pseg = seg; pseg < end_seg; pseg++)
				if (ol[pseg].aligned) break;
			/* shrinkage is local to an aligned series of words */
			if (pseg < end_seg) ol[pseg].width +=
				ol[seg].width - vwidth;
			ol[seg].width = vwidth;
		}
	    }
	}

	/* Now calculate width to be added for justification. */

	total_width = left_indent;
	#ifdef TABDEBUG
		fprintf(stderr, "line %d widths: indent %d\n", pline, left_indent);
	#endif
	for (seg = 0; seg < end_seg; seg++) {
	#ifdef TABDEBUG
		fprintf(stderr, " %d,l%d[%d]", ol[seg].width, ol[seg].len, seg);
		if (ol[seg].aligned) fprintf(stderr," aligned");
		if (ol[seg].type == TEXT_SEG) {
			fprintf(stderr," text [%c%c...]",
				*(ol[seg].s), *(ol[seg].s + 1));
		}
		fprintf(stderr,"\n");
	#endif
		if (ol[seg].type == TEXT_SEG ||
		    ol[seg].type == HRULE_SEG)
			total_width += ol[seg].width;
		if (ol[seg].type == CTL_SEG && *(ol[seg].s)==KERN_ESC)
			total_width += ol[seg].width;
	}
	#ifdef TABDEBUG
		fprintf(stderr," = %d (target = %d\n", total_width, target);
	#endif

	if (glue > 0 && target != 0) {
		more_glue =  (target - total_width) / glue;
printf("%% Remark: justified line to %.1fpt needs %d more width; adding %d to %d spcs.\n",
72.0 * target / ps_pitch_sp, target - total_width, more_glue, glue);
		target -= total_width;
		target -= more_glue * glue;
		/*if (target < 0) target = 0;*/
	}
	else more_glue = 0;

	/* Time to print. */

	curr_glue = 0;
	last_a = curr_a;
	this_x = curr_x + left_indent;
	for (seg = 0; seg < end_seg; seg++) {
		int at = ol[seg].at;
		int vflag;

		if (at != last_a && at != BOX) {
			sub = super = 0;
			if (at == SUB || at == IPASUB)
				sub = ps_baselineskip10 / 40;
			if (at == SUPER || at == IPASUPER)
				super = ps_baselineskip10 / 40;
			setfont(at);
			last_a = at;
		}

		q = ol[seg].s;
		command = *q;

		if ( 72.0*(this_x + ol[seg].width)/ps_pitch_sp >= Paper_Edge )
			return;
 
		if (command != NODE_ESC && ol[seg].treelen >= 0)
			move_to(this_x, curr_y + super - sub);

		ol[seg].mid += this_x;

		if (at == BOX) {
			makeboxes(q, ol[seg].len, this_x, curr_y);
			this_x += ol[seg].width;
			continue;
		}
		if (command == HRULE_ESC) {
			makerule(*(q+1), ol[seg].width);
			this_x += ol[seg].width;
			continue;
		}
		if (command == KERN_ESC) {
			this_x += ol[seg].width;
			continue;
		}
		if (command == PS_ESC) {
			putchar(' ');
			for (p = ol[seg].s + 2; p < ol[seg+1].s - 2; p++)
				putchar(*p);
			if (*p >= ' ') putchar(*p);
			putchar('\n');
			continue;
		}
		if (command == NODE_ESC) {
			int treeid, tlen, tcol, ref, ref2, pseg;
			int end1, end2, lbear;
			tcol = *(q+1) - 1;
			tlen = *(q+2) - 1;
			lbear = *(q+3) - 1;
			arg2 = *(q+4) - 1;
			arg3 = *(q+5) - 1;
			treeid = arg2 * 100 + arg3;

			if (seg && ol[seg-1].type == TEXT_SEG && ol[seg-1].treelen >= 0) {
				ref2 = ol[seg-1].mid;
				end2 = ref2 + ol[seg-1].nwidth / 2;
			}
			else if (seg) end2 = ref2 = ol[seg-1].mid;
			else end2 = ref2 = 0;
			/* ref2 from last word in node name */
			pseg = seg;
			while (pseg > 0 && ol[pseg-1].type == TEXT_SEG)
				pseg--;
			/* pseg is first word in node name */
			if (ol[pseg].type == TEXT_SEG && ol[seg-1].treelen >= 0) {
				ref = ol[pseg].mid;
				end1 = ref - ol[pseg].nwidth / 2;
			}
			else end1 = ref = ol[pseg].mid;
			/* make ref the midpoint of the node name: */
			printf(" %% NB: seg=%d tlen=%d, col=%d, ref=%d, ref2=%d, left bearing=%d\n", seg,
			 tlen, SP(tcol)+curr_x, ref, ref2, lbear);
			ref = (ref + ref2)/2;
			if (ref != ref2 && end1 && end2) ref = (end1 + end2)/2;
			if (!ref) ref = SP(tcol) + curr_x;
			if (!end1) end1 = ref;
			if (!end2) end2 = ref;
			printf(" %% DEFINE node %d, hpos=%d\n", treeid, ref);
			node_position(treeid, end1, end2, ref, curr_y);
			continue;
		}

		if (ol[seg].type != TEXT_SEG) continue;

		if (ol[seg].treelen < 0) {
			this_x += ol[seg].width;
			continue;
		}

		vflag = ol[seg].vkerns;
		while (ol[seg].kerns) {
		    int k, nk, v, nv, tv, sk, ns = 0;
		    int litaflag = FALSE;
		    int litbflag = FALSE;
		    unsigned char a, b, oldb, at2;
		    p = ol[seg+1].s - 1;
		    q = ol[seg+1].a - 1;
		    if (p <= ol[seg].s) break;
		    b = *p--;
		    at2 = *q--;
		    if (*p == LIT_ESC) {
			litbflag = TRUE;
		        if (p <= ol[seg].s) break;
			p--; q--;
			if (p >= ol[seg].s && b == '\t' && *p == LIT_ESC) {
				p--; q--;
				b = LIT_ESC;
			}
		    }
		    k = -10000;
		    oldb = 0;
		    tv = 0;
		    for ( ; p >= ol[seg].s; p--,q--) {
			a = *p;
			if (!litaflag && p > ol[seg].s) {
				if ( *(p - 1) == LIT_ESC ) litaflag = TRUE;
			}
			if ((a > ' ' || litaflag) && (b > ' ' || litbflag)) {
			    nk = kern(at, a, b, &v, &sk, &ns, at2);
			    nk += sk;
			    nv = v;
			    if (k > -10000) {
			        int junk;
				/* get to end of char after printing dia. */
			        k += ns;
				/* add kern between chars before and after dia.*/
			        if (sk && oldb)
			           k += kern(at, a, oldb, &junk, &junk, &junk, at);
			        if (!k) printf(" 0");
			        else printf(" %.2f", 72.0 * k / (float)ps_pitch_sp);
				if (vflag) {
				   tv -= nv;
			           if (!tv) printf(" 0");
			           else printf(" %.2f", 72.0 * tv / (float)ps_pitch_sp);
				}
			    }
			    if (sk) oldb = 0;
			    else oldb = b;
			    k = nk;
			    tv = nv;
			}
			b = a;
			at2 = *q;
			litbflag = litaflag;
			if (litaflag) { p--; q--; }
			litaflag = FALSE;
		    }
		    if (k > -10000) {
		        if (!k) printf(" 0");
		        else printf(" %.2f", 72.0 * k / (float)ps_pitch_sp);
		        if (vflag) {
		           if (!tv) printf(" 0");
		           else printf(" %.2f", 72.0 * tv / (float)ps_pitch_sp);
		        }
		    }
		    break;
		}

		printf(" (");
		for (p = ol[seg].s; p < ol[seg+1].s; p++) {
			if (*p == LIT_ESC && p < ol[seg+1].s - 1) {
				p++;
				printf("\\%03o", *p);
				continue;
			}			
			if (*p > 126) printf("\\%03o", *p);
			else if (*p > ' ') {
				if (*p == '(' || *p == ')' || *p == '\\')
					putchar('\\');
				putchar(*p);
			}
			else if (*p == ' ' && !ol[seg].aligned &&
					seg > last_aligned) {
				curr_glue += more_glue;
				if (target > 0) {
					curr_glue++;
					target--;
				}
				else if (target < 0 /*&& curr_glue > 0*/) {
					curr_glue--;
					target++;
				}
			}
		}
		if (ol[seg].vkerns) printf(") KV\n");
		else if (ol[seg].kerns) printf(") K\n");
		else printf(") s\n");
		this_x += ol[seg].width + curr_glue;
		curr_glue = 0;
	}
	if (more_glue) {
	    printf("%% NB: end is at %d (col %d) : [", this_x, COL(this_x));
	    move_to(this_x, curr_y + super - sub);
	    printf("]\n");
	}
	if (zflag) printf("%% penalty = %d\n", penalty_line[pline]);
}
